@@ -1,11 +1,11 @@
 /*
  * Copyright 1986, Larry Wall
- * 
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following condition is met:
  * 1. Redistributions of source code must retain the above copyright notice,
  * this condition and the following disclaimer.
- * 
+ *
  * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
  * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
@@ -122,14 +122,14 @@ filename(const char *at, char **name)
 }
 
 static const struct got_error *
-find_patch(FILE *fp)
+find_patch(int *done, FILE *fp)
 {
 	const struct got_error *err = NULL;
 	char	*old = NULL, *new = NULL;
 	char	*line = NULL;
 	size_t	 linesize = 0;
 	ssize_t	 linelen;
-	int	 create, git = 0;
+	int	 create, rename = 0, git = 0;
 
 	while ((linelen = getline(&line, &linesize, fp)) != -1) {
 		/*
@@ -140,15 +140,34 @@ find_patch(FILE *fp)
 		if (!strncmp(line, "--- ", 4)) {
 			free(old);
 			err = filename(line+4, &old);
+		} else if (rename && !strncmp(line, "rename from ", 12)) {
+			free(old);
+			err = filename(line+12, &old);
 		} else if (!strncmp(line, "+++ ", 4)) {
 			free(new);
 			err = filename(line+4, &new);
-		} else if (!strncmp(line, "diff --git a/", 13))
+		} else if (rename && !strncmp(line, "rename to ", 10)) {
+			free(new);
+			err = filename(line + 10, &new);
+		} else if (git && !strncmp(line, "similarity index 100%", 21))
+			rename = 1;
+		else if (!strncmp(line, "diff --git a/", 13))
 			git = 1;
 
 		if (err)
 			break;
 
+		/*
+		 * Git-style diffs with "similarity index 100%" don't
+		 * have any hunks and ends with the "rename to foobar"
+		 * line.
+		 */
+		if (rename && old != NULL && new != NULL) {
+			*done = 1;
+			err = send_patch(old, new, git);
+			break;
+		}
+
 		if (!strncmp(line, "@@ -", 4)) {
 			create = !strncmp(line+4, "0,0", 3);
 			if ((old == NULL && new == NULL) ||
@@ -199,13 +218,12 @@ strtolnum(char **str, long *n)
 }
 
 static const struct got_error *
-parse_hdr(char *s, int *ok, struct got_imsg_patch_hunk *hdr)
+parse_hdr(char *s, int *done, struct got_imsg_patch_hunk *hdr)
 {
 	static const struct got_error *err = NULL;
 
-	*ok = 1;
 	if (strncmp(s, "@@ -", 4)) {
-		*ok = 0;
+		*done = 1;
 		return NULL;
 	}
 
@@ -313,7 +331,7 @@ peek_special_line(FILE *fp)
 }
 
 static const struct got_error *
-parse_hunk(FILE *fp, int *ok)
+parse_hunk(FILE *fp, int *done)
 {
 	static const struct got_error *err = NULL;
 	struct got_imsg_patch_hunk hdr;
@@ -324,14 +342,14 @@ parse_hunk(FILE *fp, int *ok)
 
 	linelen = getline(&line, &linesize, fp);
 	if (linelen == -1) {
-		*ok = 0;
+		*done = 1;
 		goto done;
 	}
 
-	err = parse_hdr(line, ok, &hdr);
+	err = parse_hdr(line, done, &hdr);
 	if (err)
 		goto done;
-	if (!*ok) {
+	if (*done) {
 		if (fseek(fp, linelen * -1, SEEK_CUR) == -1)
 			err = got_error_from_errno("fseek");
 		goto done;
@@ -350,7 +368,7 @@ parse_hunk(FILE *fp, int *ok)
 
 			/* trailing newlines may be chopped */
 			if (leftold < 3 && leftnew < 3) {
-				*ok = 0;
+				*done = 1;
 				break;
 			}
 
@@ -411,7 +429,7 @@ read_patch(struct imsgbuf *ibuf, int fd)
 {
 	const struct got_error *err = NULL;
 	FILE *fp;
-	int ok, patch_found = 0;
+	int patch_found = 0;
 
 	if ((fp = fdopen(fd, "r")) == NULL) {
 		err = got_error_from_errno("fdopen");
@@ -420,22 +438,23 @@ read_patch(struct imsgbuf *ibuf, int fd)
 	}
 
 	while (!feof(fp)) {
-		err = find_patch(fp);
+		int done = 0;
+
+		err = find_patch(&done, fp);
 		if (err)
 			goto done;
 
 		patch_found = 1;
-		for (;;) {
-			err = parse_hunk(fp, &ok);
+
+		while (!done) {
+			err = parse_hunk(fp, &done);
 			if (err)
 				goto done;
-			if (!ok) {
-				err = send_patch_done();
-				if (err)
-					goto done;
-				break;
-			}
 		}
+
+		err = send_patch_done();
+		if (err)
+			goto done;
 	}
 
 done:
